@@ -265,7 +265,10 @@
                       :original-form form
                       :format-args args
                       :context src-context
-                      :file-name (file-info-truename file-info)
+                      :file-name (if (symbolp (file-info-truename file-info)) ; :LISP or :STREAM
+                                     ;; (pathname will be NIL in those two cases)
+                                     (file-info-truename file-info)
+                                     (file-info-pathname file-info))
                       :file-position
                       (nth-value 1 (find-source-root tlf *source-info*))
                       :path path
@@ -434,6 +437,7 @@ a STYLE-WARNING (or any more serious condition)."))
   (:documentation
    "A condition type signalled when the compiler deletes code that the user
 has written, having proved that it is unreachable."))
+(define-condition unknown-typep-note (simple-compiler-note) ())
 
 (define-condition compiler-macro-application-missed-warning
     (style-warning)
