@@ -222,13 +222,15 @@
   (:args (symbol :scs (descriptor-reg)))
   (:results (res :scs (any-reg)))
   (:result-types positive-fixnum)
+  (:args-var args)
   (:generator 2
     ;; The symbol-hash slot of NIL holds NIL because it is also the
-    ;; cdr slot, so we have to strip off the two low bits to make sure
+    ;; car slot, so we have to strip off the two low bits to make sure
     ;; it is a fixnum.  The lowtag selection magic that is required to
     ;; ensure this is explained in the comment in objdef.lisp
     (loadw res symbol symbol-hash-slot other-pointer-lowtag)
-    (inst clrrwi res res n-fixnum-tag-bits)))
+    (unless (not-nil-tn-ref-p args)
+      (inst clrrwi res res n-fixnum-tag-bits))))
 
 ;;;; Fdefinition (fdefn) objects.
 
@@ -264,9 +266,9 @@
       (inst cmpwi type simple-fun-widetag)
       ;;(inst mr lip function)
       (inst addi lip function
-            (- (ash simple-fun-code-offset word-shift) fun-pointer-lowtag))
+            (- (ash simple-fun-insts-offset word-shift) fun-pointer-lowtag))
       (inst beq normal-fn)
-      (inst lr lip (make-fixup 'closure-tramp :assembly-routine))
+      (load-asm-rtn-addr lip 'closure-tramp)
       (emit-label normal-fn)
       (storew lip fdefn fdefn-raw-addr-slot other-pointer-lowtag)
       (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
@@ -280,7 +282,7 @@
   (:results (result :scs (descriptor-reg)))
   (:generator 38
     (storew null-tn fdefn fdefn-fun-slot other-pointer-lowtag)
-    (inst lr temp (make-fixup 'undefined-tramp :assembly-routine))
+    (load-asm-rtn-addr temp 'undefined-tramp)
     (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)
     (move result fdefn)))
 
@@ -296,53 +298,10 @@
 (define-vop (dynbind)
   (:args (val :scs (any-reg descriptor-reg))
          (symbol :scs (descriptor-reg)))
-  (:temporary (:sc non-descriptor-reg :offset nl3-offset) pa-flag)
   (:temporary (:scs (descriptor-reg)) temp tls-index)
   (:generator 5
      (loadw tls-index symbol symbol-tls-index-slot other-pointer-lowtag)
-     (inst cmpwi tls-index 0)
-     (inst bne TLS-VALID)
-
-     ;; No TLS slot allocated, so allocate one.
-     ;; FIXME: this is a ridiculous number of instructions to emit inline.
-     (pseudo-atomic (pa-flag)
-       (without-scheduling ()
-         (assemble ()
-           (inst li temp (+ (static-symbol-offset '*tls-index-lock*)
-                            (ash symbol-value-slot word-shift)
-                            (- other-pointer-lowtag)))
-           OBTAIN-LOCK
-           (inst lwarx tls-index null-tn temp)
-           (inst cmpwi tls-index 0)
-           (inst bne OBTAIN-LOCK)
-           (inst stwcx. thread-base-tn null-tn temp)
-           (inst bne OBTAIN-LOCK)
-           (inst isync)
-
-           ;; Check to see if the TLS index was set while we were waiting.
-           (loadw tls-index symbol symbol-tls-index-slot other-pointer-lowtag)
-           (inst cmpwi tls-index 0)
-           (inst bne RELEASE-LOCK)
-
-           (load-symbol-value tls-index *free-tls-index*)
-           ;; FIXME: Check for TLS index overflow.
-           (inst addi tls-index tls-index n-word-bytes)
-           (store-symbol-value tls-index *free-tls-index*)
-           (inst addi tls-index tls-index (- n-word-bytes))
-           (storew tls-index symbol symbol-tls-index-slot other-pointer-lowtag)
-
-           ;; The sync instruction doesn't need to happen if we branch
-           ;; directly to RELEASE-LOCK as we didn't do any stores in that
-           ;; case.
-           (inst sync)
-           RELEASE-LOCK
-           (inst stwx zero-tn null-tn temp)
-
-           ;; temp is a boxed register, but we've been storing crap in it.
-           ;; fix it before we leave pseudo-atomic.
-           (inst li temp 0))))
-
-     TLS-VALID
+     (inst twi :eq tls-index 0)
      (inst lwzx temp thread-base-tn tls-index)
      (inst addi bsp-tn bsp-tn (* binding-size n-word-bytes))
      (storew temp bsp-tn (- binding-value-slot binding-size))
@@ -427,7 +386,7 @@
   (:variant funcallable-instance-info-offset fun-pointer-lowtag)
   (:translate %funcallable-instance-info))
 
-(define-vop (set-funcallable-instance-info word-index-set)
+(define-vop (set-funcallable-instance-info word-index-set-nr)
   (:variant funcallable-instance-info-offset fun-pointer-lowtag)
   (:translate %set-funcallable-instance-info))
 
@@ -463,18 +422,15 @@
 
 ;;;; Instance hackery:
 
-(define-vop (instance-length)
+(define-vop ()
   (:policy :fast-safe)
   (:translate %instance-length)
   (:args (struct :scs (descriptor-reg)))
-  (:temporary (:scs (non-descriptor-reg)) temp)
   (:results (res :scs (unsigned-reg)))
   (:result-types positive-fixnum)
   (:generator 4
-    (loadw temp struct 0 instance-pointer-lowtag)
-    ;; shift right 8 and mask 15 low bits =
-    ;; rotate left 24, take bit indices 17 through 31.
-    (inst rlwinm res temp (- 32 n-widetag-bits) 17 31)))
+    (loadw res struct 0 instance-pointer-lowtag)
+    (inst srwi res res instance-length-shift)))
 
 (define-vop (instance-index-ref word-index-ref)
   (:policy :fast-safe)
@@ -482,7 +438,7 @@
   (:variant instance-slots-offset instance-pointer-lowtag)
   (:arg-types instance positive-fixnum))
 
-(define-vop (instance-index-set word-index-set)
+(define-vop (instance-index-set word-index-set-nr)
   (:policy :fast-safe)
   (:translate %instance-set)
   (:variant instance-slots-offset instance-pointer-lowtag)
@@ -493,6 +449,15 @@
   (:translate %instance-cas)
   (:variant instance-slots-offset instance-pointer-lowtag)
   (:arg-types instance tagged-num * *))
+(define-vop (%raw-instance-cas/word %instance-cas)
+  (:args (object)
+         (index)
+         (old-value :scs (unsigned-reg))
+         (new-value :scs (unsigned-reg)))
+  (:arg-types * tagged-num unsigned-num unsigned-num)
+  (:results (result :scs (unsigned-reg) :from :load))
+  (:result-types unsigned-num)
+  (:translate %raw-instance-cas/word))
 
 
 ;;;; Code object frobbing.
@@ -502,7 +467,7 @@
   (:policy :fast-safe)
   (:variant 0 other-pointer-lowtag))
 
-(define-vop (code-header-set word-index-set)
+(define-vop (code-header-set word-index-set-nr)
   (:translate code-header-set)
   (:policy :fast-safe)
   (:variant 0 other-pointer-lowtag))
@@ -517,27 +482,19 @@
 
 (macrolet ((def (suffix sc primtype)
              `(progn
-                (define-vop (,(symbolicate "RAW-INSTANCE-INIT/" suffix))
-                  (:args (object :scs (descriptor-reg))
-                         (value :scs (,sc)))
-                  (:arg-types * ,primtype)
-                  (:info index)
-                  (:generator 4 (inst stw value object (offset-for-raw-slot index))))
-                (define-vop (,(symbolicate "RAW-INSTANCE-REF/" suffix) word-index-ref)
+                (define-vop (,(symbolicate "%RAW-INSTANCE-REF/" suffix) word-index-ref)
                   (:policy :fast-safe)
                   (:translate ,(symbolicate "%RAW-INSTANCE-REF/" suffix))
                   (:variant instance-slots-offset instance-pointer-lowtag)
                   (:arg-types instance positive-fixnum)
                   (:results (value :scs (,sc)))
                   (:result-types ,primtype))
-                (define-vop (,(symbolicate "RAW-INSTANCE-SET/" suffix) word-index-set)
+                (define-vop (,(symbolicate "%RAW-INSTANCE-SET/" suffix) word-index-set-nr)
                   (:policy :fast-safe)
                   (:translate ,(symbolicate "%RAW-INSTANCE-SET/" suffix))
                   (:variant instance-slots-offset instance-pointer-lowtag)
                   (:arg-types instance positive-fixnum ,primtype)
-                  (:args (object) (index) (value :scs (,sc)))
-                  (:results (result :scs (,sc)))
-                  (:result-types ,primtype)))))
+                  (:args (object) (index) (value :scs (,sc)))))))
   (def word unsigned-reg unsigned-num)
   (def signed-word signed-reg signed-num))
 
@@ -566,15 +523,7 @@
     (inst bne LOOP)
     (inst isync)))
 
-(define-vop (raw-instance-init/single)
-  (:args (object :scs (descriptor-reg))
-         (value :scs (single-reg)))
-  (:arg-types * single-float)
-  (:info index)
-  (:generator 4
-    (inst stfs value object (offset-for-raw-slot index))))
-
-(define-vop (raw-instance-ref/single)
+(define-vop ()
   (:translate %raw-instance-ref/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -588,32 +537,20 @@
                                instance-pointer-lowtag))
     (inst lfsx value object offset)))
 
-(define-vop (raw-instance-set/single)
+(define-vop ()
   (:translate %raw-instance-set/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (single-reg) :target result))
+         (value :scs (single-reg)))
   (:arg-types * positive-fixnum single-float)
-  (:results (result :scs (single-reg)))
-  (:result-types single-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst addi offset index (- (ash instance-slots-offset word-shift)
                                instance-pointer-lowtag))
-    (inst stfsx value object offset)
-    (unless (location= result value)
-      (inst frsp result value))))
+    (inst stfsx value object offset)))
 
-(define-vop (raw-instance-init/double)
-  (:args (object :scs (descriptor-reg))
-         (value :scs (double-reg)))
-  (:arg-types * double-float)
-  (:info index)
-  (:generator 4
-    (inst stfd value object (offset-for-raw-slot index))))
-
-(define-vop (raw-instance-ref/double)
+(define-vop ()
   (:translate %raw-instance-ref/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -627,35 +564,20 @@
                                instance-pointer-lowtag))
     (inst lfdx value object offset)))
 
-(define-vop (raw-instance-set/double)
+(define-vop ()
   (:translate %raw-instance-set/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (double-reg) :target result))
+         (value :scs (double-reg)))
   (:arg-types * positive-fixnum double-float)
-  (:results (result :scs (double-reg)))
-  (:result-types double-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst addi offset index (- (ash instance-slots-offset word-shift)
                                instance-pointer-lowtag))
-    (inst stfdx value object offset)
-    (unless (location= result value)
-      (inst fmr result value))))
+    (inst stfdx value object offset)))
 
-(define-vop (raw-instance-init/complex-single)
-  (:args (object :scs (descriptor-reg))
-         (value :scs (complex-single-reg)))
-  (:arg-types * complex-single-float)
-  (:info index)
-  (:generator 4
-    (inst stfs (complex-single-reg-real-tn value)
-          object (offset-for-raw-slot index))
-    (inst stfs (complex-single-reg-imag-tn value)
-          object (offset-for-raw-slot index 1))))
-
-(define-vop (raw-instance-ref/complex-single)
+(define-vop ()
   (:translate %raw-instance-ref/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -671,43 +593,22 @@
     (inst addi offset offset n-word-bytes)
     (inst lfsx (complex-single-reg-imag-tn value) object offset)))
 
-(define-vop (raw-instance-set/complex-single)
+(define-vop ()
   (:translate %raw-instance-set/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-single-reg) :target result))
+         (value :scs (complex-single-reg)))
   (:arg-types * positive-fixnum complex-single-float)
-  (:results (result :scs (complex-single-reg)))
-  (:result-types complex-single-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst addi offset index (- (ash instance-slots-offset word-shift)
                                instance-pointer-lowtag))
-    (let ((value-real (complex-single-reg-real-tn value))
-          (result-real (complex-single-reg-real-tn result)))
-      (inst stfsx value-real object offset)
-      (unless (location= result-real value-real)
-        (inst frsp result-real value-real)))
+    (inst stfsx (complex-single-reg-real-tn value) object offset)
     (inst addi offset offset n-word-bytes)
-    (let ((value-imag (complex-single-reg-imag-tn value))
-          (result-imag (complex-single-reg-imag-tn result)))
-      (inst stfsx value-imag object offset)
-      (unless (location= result-imag value-imag)
-        (inst frsp result-imag value-imag)))))
+    (inst stfsx (complex-single-reg-imag-tn value) object offset)))
 
-(define-vop (raw-instance-init/complex-double)
-  (:args (object :scs (descriptor-reg))
-         (value :scs (complex-double-reg)))
-  (:arg-types * complex-double-float)
-  (:info index)
-  (:generator 4
-    (inst stfd (complex-single-reg-real-tn value)
-          object (offset-for-raw-slot index))
-    (inst stfd (complex-double-reg-imag-tn value)
-          object (offset-for-raw-slot index 2))))
-
-(define-vop (raw-instance-ref/complex-double)
+(define-vop ()
   (:translate %raw-instance-ref/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -723,27 +624,17 @@
     (inst addi offset offset (* 2 n-word-bytes))
     (inst lfdx (complex-double-reg-imag-tn value) object offset)))
 
-(define-vop (raw-instance-set/complex-double)
+(define-vop ()
   (:translate %raw-instance-set/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-double-reg) :target result))
+         (value :scs (complex-double-reg)))
   (:arg-types * positive-fixnum complex-double-float)
-  (:results (result :scs (complex-double-reg)))
-  (:result-types complex-double-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst addi offset index (- (ash instance-slots-offset word-shift)
                                instance-pointer-lowtag))
-    (let ((value-real (complex-double-reg-real-tn value))
-          (result-real (complex-double-reg-real-tn result)))
-      (inst stfdx value-real object offset)
-      (unless (location= result-real value-real)
-        (inst fmr result-real value-real)))
+    (inst stfdx (complex-double-reg-real-tn value) object offset)
     (inst addi offset offset (* 2 n-word-bytes))
-    (let ((value-imag (complex-double-reg-imag-tn value))
-          (result-imag (complex-double-reg-imag-tn result)))
-      (inst stfdx value-imag object offset)
-      (unless (location= result-imag value-imag)
-        (inst fmr result-imag value-imag)))))
+    (inst stfdx (complex-double-reg-imag-tn value) object offset)))
