@@ -33,16 +33,16 @@
 #include "runtime.h"
 #ifndef LISP_FEATURE_WIN32
 #include <sched.h>
+#else
+#include <shellapi.h>
 #endif
 #include <errno.h>
 #include <locale.h>
 #include <limits.h>
 
-#if defined(SVR4) || defined(__linux__)
 #include <time.h>
-#endif
 
-#if !(defined(LISP_FEATURE_WIN32) && defined(LISP_FEATURE_SB_THREAD))
+#ifndef LISP_FEATURE_WIN32
 #include "signal.h"
 #endif
 
@@ -64,45 +64,12 @@
 #include "genesis/static-symbols.h"
 #include "genesis/symbol.h"
 
+struct timespec lisp_init_time;
 
-#ifdef irix
-#include <string.h>
-#include "interr.h"
-#endif
-
-#ifdef SBCL_PREFIX
-char *sbcl_home = SBCL_PREFIX"/lib/sbcl/";
-#else
 static char libpath[] = "../lib/sbcl";
-char *sbcl_home;
-#endif
+char *sbcl_runtime_home;
+char *sbcl_runtime;
 
-#ifdef LISP_FEATURE_HPUX
-extern void *return_from_lisp_stub;
-#include "genesis/closure.h"
-#include "genesis/simple-fun.h"
-#endif
-
-
-/* SIGINT handler that invokes the monitor (for when Lisp isn't up to it) */
-static void
-sigint_handler(int __attribute__((unused)) signal,
-               siginfo_t __attribute__((unused)) *info,
-               os_context_t *context)
-{
-    lose("\nSIGINT hit at 0x%08lX\n",
-         (unsigned long) *os_context_pc_addr(context));
-}
-
-/* (This is not static, because we want to be able to call it from
- * Lisp land.) */
-void
-sigint_init(void)
-{
-    SHOW("entering sigint_init()");
-    install_handler(SIGINT, sigint_handler, 1);
-    SHOW("leaving sigint_init()");
-}
 
 /*
  * helper functions for dealing with command line args
@@ -113,7 +80,7 @@ successful_malloc(size_t size)
 {
     void* result = malloc(size);
     if (0 == result) {
-        lose("malloc failure\n");
+        lose("malloc failure");
     } else {
         return result;
     }
@@ -223,15 +190,7 @@ SBCL is free software, provided as is, with absolutely no warranty.\n\
 It is mostly in the public domain; some portions are provided under\n\
 BSD-style licenses.  See the CREDITS and COPYING files in the\n\
 distribution for more information.\n\
-"
-#ifdef LISP_FEATURE_WIN32
-"\n\
-WARNING: the Windows port is fragile, particularly for multithreaded\n\
-code.  Unfortunately, the development team currently lacks the time\n\
-and resources this platform demands.\n\
-"
-#endif
-, SBCL_VERSION_STRING);
+", SBCL_VERSION_STRING);
 }
 
 /* Look for a core file to load, first in the directory named by the
@@ -244,22 +203,39 @@ search_for_core ()
     char *lookhere;
     char *stem = "/sbcl.core";
     char *core;
+    struct stat filename_stat;
 
-    if (!(env_sbcl_home && *env_sbcl_home))
-      env_sbcl_home = sbcl_home;
+    if (!(env_sbcl_home && *env_sbcl_home) ||
+        stat(env_sbcl_home, &filename_stat))
+        env_sbcl_home = sbcl_runtime_home;
     lookhere = (char *) calloc(strlen(env_sbcl_home) +
+                               strlen(libpath) +
                                strlen(stem) +
                                1,
                                sizeof(char));
-    sprintf(lookhere, "%s%s", env_sbcl_home, stem);
+    sprintf(lookhere, "%s%s%s", env_sbcl_home, libpath, stem);
     core = copied_existing_filename_or_null(lookhere);
 
-    if (!core) {
-        lose("can't find core file at %s\n", lookhere);
+    if (core) {
+        free(lookhere);
+    } else {
+        free(lookhere);
+        core = copied_existing_filename_or_null ("sbcl.core");
+        if (!core) {
+            lookhere = (char *) calloc(strlen(env_sbcl_home) +
+                                       strlen(stem) +
+                                       1,
+                                       sizeof(char));
+            sprintf(lookhere, "%s%s", env_sbcl_home, stem);
+            core = copied_existing_filename_or_null (lookhere);
+            if (core) {
+                free(lookhere);
+            } else {
+                return NULL;
+            }
+        }
     }
 
-    free(lookhere);
-
     return core;
 }
 
@@ -364,57 +340,15 @@ parse_size_arg(char *arg, char *arg_name)
   return res;
 }
 
-char **posix_argv;
+#ifdef LISP_FEATURE_WIN32
+    wchar_t
+#else
+    char
+#endif
+    **posix_argv;
+
 char *core_string;
 
-char *saved_runtime_path = NULL;
-#if defined(LISP_FEATURE_WIN32) && defined(LISP_FEATURE_SB_THREAD)
-void pthreads_win32_init();
-#endif
-
-static void print_locale_variable(const char *name)
-{
-  char *value = getenv(name);
-
-  if (value) {
-    fprintf(stderr, "\n  %s=%s", name, value);
-  }
-}
-
-static void setup_locale()
-{
-  if(setlocale(LC_ALL, "") == NULL) {
-#ifndef LISP_FEATURE_WIN32
-
-    fprintf(stderr, "WARNING: Setting locale failed.\n");
-    fprintf(stderr, "  Check the following variables for correct values:");
-
-    if (setlocale(LC_CTYPE, "") == NULL) {
-      print_locale_variable("LC_ALL");
-      print_locale_variable("LC_CTYPE");
-      print_locale_variable("LANG");
-    }
-
-    if (setlocale(LC_MESSAGES, "") == NULL) {
-      print_locale_variable("LC_MESSAGES");
-    }
-    if (setlocale(LC_COLLATE, "") == NULL) {
-      print_locale_variable("LC_COLLATE");
-    }
-    if (setlocale(LC_MONETARY, "") == NULL) {
-      print_locale_variable("LC_MONETARY");
-    }
-    if (setlocale(LC_NUMERIC, "") == NULL) {
-      print_locale_variable("LC_NUMERIC");
-    }
-    if (setlocale(LC_TIME, "") == NULL) {
-      print_locale_variable("LC_TIME");
-    }
-    fprintf(stderr, "\n");
-
-#endif
-  }
-}
 static void print_environment(int argc, char *argv[])
 {
     int n = 0;
@@ -430,6 +364,33 @@ static void print_environment(int argc, char *argv[])
         ++n;
     }
 }
+
+char * sb_realpath (char *);
+char *dir_name(char *path) {
+    if (path == NULL)
+        return NULL;
+
+    char* result;
+    char slashchar =
+#ifdef LISP_FEATURE_WIN32
+        '\\';
+#else
+    '/';
+#endif
+
+    char *slash = strrchr(path, slashchar);
+
+    if (slash) {
+        int prefixlen = slash - path + 1; // keep the slash in the prefix
+        result = successful_malloc(prefixlen + 1);
+        memcpy(result, path, prefixlen);
+        result[prefixlen] = 0;
+        return result;
+    } else {
+        return NULL;
+    }
+}
+
 
 extern void write_protect_immobile_space();
 struct lisp_startup_options lisp_startup_options;
@@ -440,13 +401,27 @@ sbcl_main(int argc, char *argv[], char *envp[])
     /* Exception handling support structure. Evil Win32 hack. */
     struct lisp_exception_frame exception_frame;
 #endif
+#ifdef LISP_FEATURE_UNIX
+    clock_gettime(
+#ifdef LISP_FEATURE_LINUX
+        CLOCK_MONOTONIC_COARSE
+#else
+        CLOCK_MONOTONIC
+#endif
+        , &lisp_init_time);
+#endif
 
     /* the name of the core file we're to execute. Note that this is
      * a malloc'ed string which should be freed eventually. */
     char *core = 0;
-    char **sbcl_argv = 0;
+
+#ifdef LISP_FEATURE_WIN32
+    wchar_t
+#else
+        char
+#endif
+        **sbcl_argv = 0;
     os_vm_offset_t embedded_core_offset = 0;
-    char *runtime_path = 0;
 
     /* other command line options */
     boolean end_runtime_options = 0;
@@ -461,47 +436,49 @@ sbcl_main(int argc, char *argv[], char *envp[])
 
     lispobj initial_function;
     int merge_core_pages = -1;
-    struct memsize_options memsize_options = {0, 0, 0};
+    struct memsize_options memsize_options;
+    memsize_options.present_in_core = 0;
 
     boolean have_hardwired_spaces = os_preinit(argv, envp);
-#if defined(LISP_FEATURE_WIN32) && defined(LISP_FEATURE_SB_THREAD)
-    pthreads_win32_init();
-#endif
 
     interrupt_init();
+#ifdef LISP_FEATURE_UNIX
+    /* Not sure why anyone sends signals to this process so early.
+     * But win32 models the signal mask as part of 'struct thread'
+     * which doesn't exist yet, so don't do this */
     block_blockable_signals(0);
-
-    /* Save the argv[0] derived runtime path in case
-     * os_get_runtime_executable_path(1) isn't able to get an
-     * externally-usable path later on. */
-    saved_runtime_path = search_for_executable(argv[0]);
+#endif
 
     /* Check early to see if this executable has an embedded core,
      * which also populates runtime_options if the core has runtime
      * options */
-    runtime_path = os_get_runtime_executable_path(0);
-    if (runtime_path || saved_runtime_path) {
-        os_vm_offset_t offset = search_for_embedded_core(
-            runtime_path ? runtime_path : saved_runtime_path,
-            &memsize_options);
+    if (!(sbcl_runtime = os_get_runtime_executable_path()))
+        sbcl_runtime = search_for_executable(argv[0]);
+
+    if (!(sbcl_runtime_home = dir_name(argv[0])))
+      if (!(sbcl_runtime_home = dir_name(sbcl_runtime)))
+        sbcl_runtime_home = libpath;
+
+    if (sbcl_runtime) {
+        os_vm_offset_t offset = search_for_embedded_core(sbcl_runtime, &memsize_options);
         if (offset != -1) {
             embedded_core_offset = offset;
-            core = (runtime_path ? runtime_path :
-                    copied_string(saved_runtime_path));
-        } else {
-            if (runtime_path)
-                free(runtime_path);
+            core = sbcl_runtime;
         }
     }
 
-
     /* Parse our part of the command line (aka "runtime options"),
      * stripping out those options that we handle. */
     if (memsize_options.present_in_core) {
         dynamic_space_size = memsize_options.dynamic_space_size;
         thread_control_stack_size = memsize_options.thread_control_stack_size;
         dynamic_values_bytes = memsize_options.thread_tls_bytes;
+#ifndef LISP_FEATURE_WIN32
         sbcl_argv = argv;
+#else
+        int wargc;
+        sbcl_argv = CommandLineToArgvW(GetCommandLineW(), &wargc);
+#endif
     } else {
         int argi = 1;
 
@@ -522,11 +499,11 @@ sbcl_main(int argc, char *argv[], char *envp[])
                 ++argi;
             } else if (0 == strcmp(arg, "--core")) {
                 if (core) {
-                    lose("more than one core file specified\n");
+                    lose("more than one core file specified");
                 } else {
                     ++argi;
                     if (argi >= argc) {
-                        lose("missing filename for --core argument\n");
+                        lose("missing filename for --core argument");
                     }
                     core = copied_string(argv[argi]);
                     ++argi;
@@ -608,6 +585,7 @@ sbcl_main(int argc, char *argv[], char *envp[])
         {
             char *argi0 = argv[argi];
             int argj = 1;
+#ifndef LISP_FEATURE_WIN32
             /* (argc - argi) for the arguments, one for the binary,
                and one for the terminating NULL. */
             sbcl_argv = successful_malloc((2 + argc - argi) * sizeof(char *));
@@ -621,10 +599,28 @@ sbcl_main(int argc, char *argv[], char *envp[])
                  * error. */
                 if (!end_runtime_options &&
                     0 == strcmp(arg, "--end-runtime-options")) {
-                    lose("bad runtime option \"%s\"\n", argi0);
+                    lose("bad runtime option \"%s\"", argi0);
                 }
                 sbcl_argv[argj++] = arg;
             }
+#else
+            /* The runtime options are processed as chars above, which may
+             * not always work but may be good enough for now, as it
+             * has been for a long time. */
+            int wargc;
+            wchar_t** wargv;
+            wargv = CommandLineToArgvW(GetCommandLineW(), &wargc);
+            sbcl_argv = successful_malloc((2 + wargc - argi) * sizeof(wchar_t *));
+            sbcl_argv[0] = wargv[0];
+            while (argi < wargc) {
+                wchar_t *warg = wargv[argi++];
+                if (!end_runtime_options &&
+                    0 == wcscmp(warg, L"--end-runtime-options")) {
+                    lose("bad runtime option \"%s\"", argi0);
+                }
+                sbcl_argv[argj++] = warg;
+            }
+#endif
             sbcl_argv[argj] = 0;
         }
     }
@@ -637,51 +633,35 @@ sbcl_main(int argc, char *argv[], char *envp[])
 #endif
     thread_control_stack_size &= ~(sword_t)(CONTROL_STACK_ALIGNMENT_BYTES-1);
 
-    /* KLUDGE: os_vm_page_size is set by os_init(), and on some
-     * systems (e.g. Alpha) arch_init() needs need os_vm_page_size, so
-     * it must follow os_init(). -- WHN 2000-01-26 */
     os_init(argv, envp);
     if (debug_environment_p) {
         print_environment(argc, argv);
     }
     dyndebug_init();
-    arch_init();
-    allocate_spaces(have_hardwired_spaces);
+    // FIXME: if the 'have' flag is 0 and you've disabled disabling of ASLR
+    // then we haven't done an exec(), nor unmapped the mappings that were obtained
+    // already obtained (if any) so it is unhelpful to try again here.
+    allocate_lisp_dynamic_space(have_hardwired_spaces);
     gc_init();
 
-    setup_locale();
-
-    #ifndef SBCL_PREFIX
-    /* If built without SBCL_PREFIX defined, then set 'sbcl_home' to
-     * "<here>/../lib/sbcl/" based on how this executable was invoked. */
-    {
-        char *exename = argv[0]; // Use as-it, not truenameified
-        char *slash = strrchr(exename, '/');
-        if (!slash) {
-            sbcl_home = libpath;
-        } else {
-            int prefixlen = slash - exename + 1; // keep the slash in the prefix
-            char *tail = exename + prefixlen - 4;
-            char *suffix = libpath;
-            sbcl_home = successful_malloc(prefixlen + sizeof libpath); // sizeof incl. nul
-            // Translate "{path/}bin/sbcl" => "{path/}lib/sbcl", otherwise
-            // "{path}/sbcl" => "{path}/../lib/sbcl" so that running "./sbcl" works
-            // if sitting in "bin".
-            if (prefixlen >= 4 && !strncmp(tail, "bin/", 4)
-                // chop "bin" only if a complete word: '/' or nothing to its left.
-                && (tail-1 < exename || tail[-1] == '/')) {
-                prefixlen -= 4; // remove "bin/"
-                suffix += 3; // don't append "../"
-            }
-            memcpy(sbcl_home, exename, prefixlen);
-            strcpy(sbcl_home+prefixlen, suffix);
-        }
-    }
-    #endif
-
     /* If no core file was specified, look for one. */
-    if (!core) {
-        core = search_for_core();
+    if (!core && !(core = search_for_core())) {
+      /* Try resolving symlinks */
+      if (sbcl_runtime) {
+        free(sbcl_runtime_home);
+        char* real = sb_realpath(sbcl_runtime);
+        if (!real)
+          goto lose;
+        sbcl_runtime_home = dir_name(real);
+        free(real);
+        if (!sbcl_runtime_home)
+          goto lose;
+        if(!(core = search_for_core()))
+          goto lose;
+      } else {
+      lose:
+        lose("Can't find sbcl.core");
+      }
     }
 
     if (embedded_core_offset)
@@ -714,10 +694,10 @@ sbcl_main(int argc, char *argv[], char *envp[])
     initial_function = load_core_file(core, embedded_core_offset,
                                       merge_core_pages);
     if (initial_function == NIL) {
-        lose("couldn't find initial function\n");
+        lose("couldn't find initial function");
     }
 
-#if defined(SVR4) || defined(__linux__) || defined(__NetBSD__)
+#if defined(SVR4) || defined(__linux__) || defined(__NetBSD__) || defined(__HAIKU__)
     tzset();
 #endif
 
@@ -734,16 +714,15 @@ sbcl_main(int argc, char *argv[], char *envp[])
      * since it was too soon earlier to handle write faults. */
     write_protect_immobile_space();
 #endif
-#ifdef LISP_FEATURE_HPUX
-    /* -1 = CLOSURE_FUN_OFFSET, 23 = SIMPLE_FUN_CODE_OFFSET, we are
-     * not in __ASSEMBLER__ so we cant reach them. */
-    return_from_lisp_stub = (void *) ((char *)*((unsigned long *)
-                 ((char *)initial_function + -1)) + 23);
-#endif
 
     arch_install_interrupt_handlers();
 #ifndef LISP_FEATURE_WIN32
     os_install_interrupt_handlers();
+# ifdef LISP_FEATURE_SB_SAFEPOINT
+    ll_install_handler(SIGURG, thruption_handler);
+# elif defined LISP_FEATURE_SB_THREAD
+    ll_install_handler(SIG_STOP_FOR_GC, sig_stop_for_gc_handler);
+# endif
 #else
 /*     wos_install_interrupt_handlers(handler); */
     wos_install_interrupt_handlers(&exception_frame);
@@ -757,7 +736,7 @@ sbcl_main(int argc, char *argv[], char *envp[])
 
     FSHOW((stderr, "/funcalling initial_function=0x%lx\n",
           (unsigned long)initial_function));
-    create_initial_thread(initial_function);
-    lose("unexpected return from initial thread in main()\n");
+    create_main_lisp_thread(initial_function);
+    lose("unexpected return from initial thread in main()");
     return 0;
 }
