@@ -16,7 +16,6 @@
 #include "globals.h"
 #include "validate.h"
 #include "os.h"
-#include "sbcl.h"
 #include "arch.h"
 #include "lispregs.h"
 #include "signal.h"
@@ -25,25 +24,20 @@
 #include "interr.h"
 #include "breakpoint.h"
 #include "thread.h"
-#include "pseudo-atomic.h"
+#include "getallocptr.h"
 #include "unaligned.h"
+#include "search.h"
 
 #include "genesis/static-symbols.h"
 #include "genesis/symbol.h"
+#include "forwarding-ptr.h"
+#include "core.h"
 
-
-#ifdef LISP_FEATURE_UD2_BREAKPOINTS
-#define UD2_INST 0x0b0f         /* UD2 */
-#define BREAKPOINT_WIDTH 2
-#else
-#ifdef LISP_FEATURE_INT4_BREAKPOINTS
-# define BREAKPOINT_INST 0xce    /* INTO */
-#else
-# define BREAKPOINT_INST 0xcc    /* INT3 */
-#endif
+#define INT3_INST 0xCC
+#define INTO_INST 0xCE
+#define UD2_INST 0x0b0f
 #define BREAKPOINT_WIDTH 1
-#endif
-unsigned int cpuid_fn1_ecx;
+
 int avx_supported = 0, avx2_supported = 0;
 
 static void cpuid(unsigned info, unsigned subinfo,
@@ -80,6 +74,7 @@ static void xgetbv(unsigned *eax, unsigned *edx)
 void tune_asm_routines_for_microarch(void)
 {
     unsigned int eax, ebx, ecx, edx;
+    unsigned int cpuid_fn1_ecx = 0;
 
     cpuid(0, 0, &eax, &ebx, &ecx, &edx);
     if (eax >= 1) { // see if we can execute basic id function 1
@@ -97,7 +92,13 @@ void tune_asm_routines_for_microarch(void)
             }
         }
     }
-    SetSymbolValue(CPUID_FN1_ECX, (lispobj)make_fixnum(cpuid_fn1_ecx), 0);
+    int our_cpu_feature_bits = 0;
+    // avx2_supported gets copied into bit 1 of *CPU-FEATURE-BITS*
+    if (avx2_supported) our_cpu_feature_bits |= 1;
+    // POPCNT = ECX bit 23, which gets copied into bit 2 in *CPU-FEATURE-BITS*
+    if (cpuid_fn1_ecx & (1<<23)) our_cpu_feature_bits |= 2;
+    SetSymbolValue(CPU_FEATURE_BITS, make_fixnum(our_cpu_feature_bits), 0);
+
     // I don't know if this works on Windows
 #ifndef _MSC_VER
     cpuid(0, 0, &eax, &ebx, &ecx, &edx);
@@ -116,6 +117,7 @@ void tune_asm_routines_for_microarch(void)
 void untune_asm_routines_for_microarch(void)
 {
     asm_routine_poke(VECTOR_FILL_T, 0x12, 0xEB); // Change JL to JMP
+    SetSymbolValue(CPU_FEATURE_BITS, 0, 0);
 }
 
 #ifndef _WIN64
@@ -150,6 +152,8 @@ context_eflags_addr(os_context_t *context)
     return &context->uc_mcontext.gregs[REG_RFL];
 #elif defined LISP_FEATURE_FREEBSD || defined(__DragonFly__)
     return &context->uc_mcontext.mc_rflags;
+#elif defined __HAIKU__
+    return &context->uc_mcontext.rflags;
 #elif defined LISP_FEATURE_DARWIN
     return CONTEXT_ADDR_FROM_STEM(rflags);
 #elif defined __OpenBSD__
@@ -179,9 +183,12 @@ void arch_skip_instruction(os_context_t *context)
         case trap_Error:
         case trap_Cerror:
             skip_internal_error(context);
-
             break;
-
+        case trap_UninitializedLoad:
+            // Skip 1 byte. We can't encode that the internal_error_nargs is 1
+            // because it is not an SC+OFFSET that follows the trap code.
+            *os_context_pc_addr(context) += 1;
+            break;
         case trap_Breakpoint:           /* not tested */
         case trap_FunEndBreakpoint: /* not tested */
             break;
@@ -217,20 +224,20 @@ arch_internal_error_arguments(os_context_t *context)
 boolean
 arch_pseudo_atomic_atomic(os_context_t __attribute__((unused)) *context)
 {
-    return get_pseudo_atomic_atomic(arch_os_get_current_thread());
+    return get_pseudo_atomic_atomic(get_sb_vm_thread());
 }
 
 void
 arch_set_pseudo_atomic_interrupted(os_context_t __attribute__((unused)) *context)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     set_pseudo_atomic_interrupted(thread);
 }
 
 void
 arch_clear_pseudo_atomic_interrupted(os_context_t __attribute__((unused)) *context)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     clear_pseudo_atomic_interrupted(thread);
 }
 
@@ -242,16 +249,12 @@ unsigned int
 arch_install_breakpoint(void *pc)
 {
     unsigned int result = *(unsigned int*)pc;
-
-#ifndef LISP_FEATURE_UD2_BREAKPOINTS
-    *(char*)pc = BREAKPOINT_INST;               /* x86 INT3       */
-    *((char*)pc+1) = trap_Breakpoint;           /* Lisp trap code */
+#ifdef LISP_FEATURE_INT4_BREAKPOINTS
+    *(char*)pc = INTO_INST;
 #else
-    *(char*)pc = UD2_INST & 0xff;
-    *((char*)pc+1) = UD2_INST >> 8;
-    *((char*)pc+2) = trap_Breakpoint;
+    *(char*)pc = INT3_INST;
 #endif
-
+    *((char*)pc+1) = trap_Breakpoint;           /* Lisp trap code */
     return result;
 }
 
@@ -260,9 +263,6 @@ arch_remove_breakpoint(void *pc, unsigned int orig_inst)
 {
     *((char *)pc) = orig_inst & 0xff;
     *((char *)pc + 1) = (orig_inst & 0xff00) >> 8;
-#if BREAKPOINT_WIDTH > 1
-    *((char *)pc + 2) = (orig_inst & 0xff0000) >> 16;
-#endif
 }
 
 /* When single stepping, single_stepping holds the original instruction
@@ -366,15 +366,30 @@ sigtrap_handler(int __attribute__((unused)) signal,
 
     /* This is just for info in case the monitor wants to print an
      * approximation. */
-    access_control_stack_pointer(arch_os_get_current_thread()) =
+    access_control_stack_pointer(get_sb_vm_thread()) =
         (lispobj *)*os_context_sp_addr(context);
 
-    /* On entry %eip points just after the INT3 byte and aims at the
+    /* On entry %rip points just after the INT3 byte and aims at the
      * 'kind' value (eg trap_Cerror). For error-trap and Cerror-trap a
      * number of bytes will follow, the first is the length of the byte
      * arguments to follow. */
     trap = *(unsigned char *)(*os_context_pc_addr(context));
-
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    if (trap == trap_UndefinedFunction) {
+        // The interrupted PC pins this fdefn. Sigtrap is delivered on the ordinary stack,
+        // not the alternate stack.
+        // (FIXME: an interior pointer to an fdefn _should_ pin it, but doesn't)
+        lispobj* fdefn = (lispobj*)(*os_context_pc_addr(context) & ~LOWTAG_MASK);
+        if (fdefn && widetag_of(fdefn) == FDEFN_WIDETAG) {
+            // Return to undefined-tramp
+            *os_context_pc_addr(context) = (uword_t)((struct fdefn*)fdefn)->raw_addr;
+            // with RAX containing the FDEFN
+            *os_context_register_addr(context,reg_RAX) =
+                make_lispobj(fdefn, OTHER_POINTER_LOWTAG);
+            return;
+        }
+    }
+#endif
     handle_trap(context, trap);
 }
 
@@ -382,23 +397,25 @@ void
 sigill_handler(int __attribute__((unused)) signal,
                siginfo_t __attribute__((unused)) *siginfo,
                os_context_t *context) {
-    /* Triggering SIGTRAP using int3 is unreliable on OS X/x86, so
-     * we need to use illegal instructions for traps.
-     */
-#if defined(LISP_FEATURE_UD2_BREAKPOINTS) && !defined(LISP_FEATURE_MACH_EXCEPTION_HANDLER)
-    if (*((unsigned short *)*os_context_pc_addr(context)) == UD2_INST) {
+    unsigned char* pc = (void*)*os_context_pc_addr(context);
+#ifndef LISP_FEATURE_MACH_EXCEPTION_HANDLER
+    if (*(unsigned short *)pc == UD2_INST) {
         *os_context_pc_addr(context) += 2;
         return sigtrap_handler(signal, siginfo, context);
     }
-#elif defined(LISP_FEATURE_INT4_BREAKPOINTS) && !defined(LISP_FEATURE_MACH_EXCEPTION_HANDLER)
-    if (*((unsigned char *)*os_context_pc_addr(context)) == BREAKPOINT_INST) {
-        *os_context_pc_addr(context) += BREAKPOINT_WIDTH;
+    // Interrupt if overflow (INTO) raises SIGILL in 64-bit mode
+    if (*(unsigned char *)pc == INTO_INST) {
+        *os_context_pc_addr(context) += 1;
         return sigtrap_handler(signal, siginfo, context);
     }
 #endif
 
     fake_foreign_function_call(context);
-    lose("Unhandled SIGILL at %p.", (void*)*os_context_pc_addr(context));
+#ifdef LISP_FEATURE_LINUX
+    extern void sb_dump_mcontext(char*,void*);
+    sb_dump_mcontext("SIGILL received", context);
+#endif
+    lose("Unhandled SIGILL at %p.", pc);
 }
 
 #ifdef X86_64_SIGFPE_FIXUP
@@ -470,21 +487,21 @@ arch_install_interrupt_handlers()
      * CL way, I hope there will at least be a comment to explain
      * why.. -- WHN 2001-06-07 */
 #if !defined(LISP_FEATURE_MACH_EXCEPTION_HANDLER) && !defined(LISP_FEATURE_WIN32)
-    undoably_install_low_level_interrupt_handler(SIGILL , sigill_handler);
-    undoably_install_low_level_interrupt_handler(SIGTRAP, sigtrap_handler);
+    ll_install_handler(SIGILL , sigill_handler);
+    ll_install_handler(SIGTRAP, sigtrap_handler);
 #endif
 
 #if defined(X86_64_SIGFPE_FIXUP) && !defined(LISP_FEATURE_WIN32)
-    undoably_install_low_level_interrupt_handler(SIGFPE, sigfpe_handler);
+    ll_install_handler(SIGFPE, sigfpe_handler);
 #endif
 
     SHOW("returning from arch_install_interrupt_handlers()");
 }
 
-#ifdef LISP_FEATURE_LINKAGE_TABLE
 void
-arch_write_linkage_table_entry(char *reloc_addr, void *target_addr, int datap)
+arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 {
+    char *reloc_addr = (char*)LINKAGE_TABLE_SPACE_START + index * LINKAGE_TABLE_ENTRY_SIZE;
     if (datap) {
         *(uword_t *)reloc_addr = (uword_t)target_addr;
         return;
@@ -495,7 +512,6 @@ arch_write_linkage_table_entry(char *reloc_addr, void *target_addr, int datap)
     reloc_addr[6] = 0x66; reloc_addr[7] = 0x90; /* 2-byte NOP */
     *(void**)(reloc_addr+8) = target_addr;
 }
-#endif
 
 /* These setup and check *both* the sse2 and x87 FPUs. While lisp code
    only uses the sse2 FPU, other code (such as libc) may use the x87 FPU.
@@ -551,7 +567,6 @@ arch_set_fp_modes(unsigned int mxcsr)
     asm ("ldmxcsr %0" : : "m" (temp));
 }
 
-#ifdef LISP_FEATURE_IMMOBILE_CODE
 /// Return the Lisp object that fdefn's raw_addr slot jumps to.
 /// This will either be:
 /// (1) a simple-fun,
@@ -560,22 +575,36 @@ arch_set_fp_modes(unsigned int mxcsr)
 /// (3) a code-component with no simple-fun within it, that makes
 ///     closures and other funcallable-instances look like simple-funs.
 /// If the fdefn jumps to the UNDEFINED-FDEFN routine, then return 0.
-lispobj virtual_fdefn_callee_lispobj(struct fdefn* fdefn, uword_t vaddr) {
-    unsigned char tagged_ptr_bias = ((unsigned char*)&fdefn->raw_addr)[7];
-    // If the pointer bias is 0, then this fdefn's raw_addr must
-    // point to an assembler routine entry point.
-    if (tagged_ptr_bias == 0)
-        return 0;
-    int32_t offs = UNALIGNED_LOAD32((char*)&fdefn->raw_addr + 1);
-    // Base the callee address off where the fdefn virtually is.
-    // Compensate for the offset of the raw_addr slot,
-    // and add 5 for the length of "JMP rel32" instruction.
-    return vaddr + offsetof(struct fdefn,raw_addr) + 5 + offs - tagged_ptr_bias;
-}
 lispobj fdefn_callee_lispobj(struct fdefn* fdefn) {
-  return virtual_fdefn_callee_lispobj(fdefn, (uword_t)fdefn);
+    lispobj* raw_addr = (lispobj*)fdefn->raw_addr;
+    if (!raw_addr || points_to_asm_code_p((lispobj)raw_addr))
+        // technically this should return the address of the code object
+        // containing asm routines, but it's fine to return 0.
+        return 0;
+    // If the object to which raw_addr points was already forwarded,
+    // this returns the "old" pointer, prior to forwarding, so that
+    // scavenging that pointer alters it. Otherwise scav_fdefn would not
+    // decide to rewrite the raw_addr slot of the fdefn.
+    // This logic is rather nasty, but I don't know what else to do.
+    lispobj word;
+    if (header_widetag(word = raw_addr[-2]) == SIMPLE_FUN_WIDETAG
+        || (word == 1 &&
+            widetag_of(native_pointer(forwarding_pointer_value(raw_addr-2)))
+            == SIMPLE_FUN_WIDETAG))
+        return make_lispobj(raw_addr - 2, FUN_POINTER_LOWTAG);
+    int widetag;
+    if ((widetag = header_widetag(word = raw_addr[-4])) == CODE_HEADER_WIDETAG
+        || (word == 1 &&
+            widetag_of(native_pointer(forwarding_pointer_value(raw_addr-4)))
+            == CODE_HEADER_WIDETAG))
+        return make_lispobj(raw_addr - 4, OTHER_POINTER_LOWTAG);
+    if (widetag == FUNCALLABLE_INSTANCE_WIDETAG
+        || (word == 1 &&
+            widetag_of(native_pointer(forwarding_pointer_value(raw_addr-4)))
+            == FUNCALLABLE_INSTANCE_WIDETAG))
+        return make_lispobj(raw_addr - 4, FUN_POINTER_LOWTAG);
+    lose("Unknown object in fdefn raw addr: %p", raw_addr);
 }
-#endif
 
 #include "genesis/vector.h"
 #define LOCK_PREFIX 0xF0
@@ -584,8 +613,12 @@ lispobj fdefn_callee_lispobj(struct fdefn* fdefn) {
 extern unsigned int alloc_profile_n_counters;
 extern unsigned int max_alloc_point_counters;
 #ifdef LISP_FEATURE_SB_THREAD
+#ifdef LISP_FEATURE_WIN32
+extern CRITICAL_SECTION alloc_profiler_lock;
+#else
 extern pthread_mutex_t alloc_profiler_lock;
 #endif
+#endif
 
 static unsigned int claim_index(int qty)
 {
@@ -601,7 +634,8 @@ static unsigned int claim_index(int qty)
     return 0; // use the overflow bin(s)
 }
 
-static boolean instrumentp(uword_t* sp, uword_t** pc, uword_t* old_word)
+static boolean NO_SANITIZE_MEMORY
+instrumentp(uword_t* sp, uword_t** pc, uword_t* old_word)
 {
     int __attribute__((unused)) ret = thread_mutex_lock(&alloc_profiler_lock);
     gc_assert(ret == 0);
@@ -674,9 +708,10 @@ allocation_tracker_counted(uword_t* sp)
         unsigned int index = claim_index(1);
         if (index == 0)
             index = 2; // reserved overflow counter for fixed-size alloc
+        uword_t disp = index * 8;
         // rewrite call into: LOCK INC QWORD PTR, [R11+n] ; opcode = 0xFF / 0
         uword_t new_inst =
-          0xF0 | (0x49 << 8) | (0xFF << 16) | (0x83L << 24) | ((index*8L) << 32);
+          0xF0 | (0x49 << 8) | (0xFF << 16) | (0x83L << 24) | (disp << 32);
         // Ensure atomicity of the write. A plain store would probably do,
         // but since this is self-modifying code, the most stringent memory
         // order is prudent.
@@ -704,18 +739,19 @@ allocation_tracker_sized(uword_t* sp)
             (modrm & 0xC0) == 0xC0 && /* register-direct mode */
             ((modrm >> 3) & 0x7) == (modrm & 0x7)) { /* same register */
         } else {
-            lose("Can't decode instruction @ pc %p\n", pc);
+            lose("Can't decode instruction @ pc %p", pc);
         }
         // rewrite call into:
         //  LOCK INC QWORD PTR, [R11+n] ; opcode = 0xFF / 0
+        uword_t disp = index * 8;
         uword_t new_inst1 =
-          0xF0 | (0x49 << 8) | (0xFF << 16) | (0x83L << 24) | ((index * 8L) << 32);
+          0xF0 | (0x49 << 8) | (0xFF << 16) | (0x83L << 24) | (disp << 32);
         //  LOCK ADD [R11+n], Rxx ; opcode = 0x01
         prefix = 0x49 | ((prefix & 1) << 2); // 'b' bit becomes 'r' bit
         modrm  = 0x83 | (modrm & (7<<3)); // copy 'reg' into new modrm byte
+        disp = (1 + index) * 8;
         uword_t new_inst2 =
-          0xF0 | (prefix << 8) | (0x01 << 16) | ((long)modrm << 24)
-            | (((1 + index) * 8L) << 32);
+          0xF0 | (prefix << 8) | (0x01 << 16) | ((long)modrm << 24) | (disp << 32);
         // Overwrite the second instruction first, because as soon as the CALL
         // opcode is changed, fallthrough to the next instruction occurs.
         if (!__sync_bool_compare_and_swap(pc+1, word_after_pc, new_inst2) ||
